
class Group:
    def __init__(self, label, order=None):
        self.label = label
        self.order = order if order else 0

    def __lt__(self, other):
        return (
            self.order < other.order
            or self.order == other.order and self.label < other.label
        )

    def __eq__(self, other):
        return self.order == other.order and self.label == other.label


Group.APP = Group('*', order=0)
Group.FILE = Group('File', order=1)
Group.EDIT = Group('Edit', order=10)
Group.VIEW = Group('View', order=20)
Group.COMMANDS = Group('Commands', order=30)
Group.WINDOW = Group('Window', order=90)
Group.HELP = Group('Help', order=100)


class Command:
    def __init__(self, action, label,
                 shortcut=None, tooltip=None, icon=None,
                 group=None, section=None, order=None):
        self.action = action
        self.label = label

        self.shortcut = shortcut
        self.tooltip = tooltip
        self.icon_id = icon

        self.group = group if group else Group.COMMANDS
        self.section = section if section else 0
        self.order = order if order else 0

        self._enabled = self.action is not None

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = value
        self._set_enabled(value)

    def _set_enabled(self, value):
        pass


GROUP_BREAK = object()
SECTION_BREAK = object()


def cmd_sort_key(value):
    return (value.group, value.section, value.order, value.label)


class CommandSet:
    def __init__(self, widget, on_change=None):
        self.widget = widget
        self._values = set()
        self.on_change = on_change

    def add(self, *values):
        if self.widget:
            self.widget.app.commands.add(*values)
        self._values.update(values)
        if self.on_change:
            self.on_change()

    def __iter__(self):
        prev_cmd = None
        for cmd in sorted(self._values, key=cmd_sort_key):
            if prev_cmd:
                if cmd.group != prev_cmd.group:
                    yield GROUP_BREAK
                elif cmd.section != prev_cmd.section:
                    yield SECTION_BREAK

            yield cmd
            prev_cmd = cmd



Black = 'BLACK'
Blue = 'BLUE'
Brown = 'BROWN'
Clear = 'CLEAR'
Cyan = 'CYAN'
DarkGray = 'DARKGRAY'
Gray = 'GRAY'
Green = 'GREEN'
LightGray = 'LIGHTGRAY'
Magenta = 'MAGENTA'
Orange = 'ORANGE'
Purple = 'PURPLE'
Red = 'RED'
White = 'WHITE'
Yellow = 'YELLOW'



from .base import Widget


class Range:
    def __init__(self, max, min):
        self.max = max
        self.min = min


class Slider(Widget):
    """
    Slider widget, displays a range of values

    :param id:          An identifier for this widget.
    :type  id:          ``str``

    :param style:       an optional style object. If no style is provided then a
                        new one will be created for the widget.
    :type style:        :class:`colosseum.CSSNode`

    :param default:     Default value of the slider
    :type default:      ``float``

    :param range:       Min and max values of the slider
    :type range:        ``Range``

    :param on_slide:    Function to execute on slide
    :type on_slide:     ``callable``

    :param enabled:     Whether user interaction is possible or not.
    :type enabled:      ``Bool``
    """

    def __init__(self, id=None, style=None, default=None, range=None, on_slide=None, enabled=True):
        super().__init__(id=id, style=style, default=default, range=range, on_slide=on_slide, enabled=enabled)

    def _configure(self, default, range, on_slide, enabled):
        self.range = range
        self.value = default
        self.on_slide = on_slide
        self.enabled = enabled

    @property
    def value(self):
        """
        :returns: The current slider value
        :rtype: ``float``
        """
        return self._get_value()

    @value.setter
    def value(self, value):
        """
        Set the value of the slider.

        :param value:       The new slider value
        :type value:        ``int or float``
        """
        _min, _max = self.range
        if value is None:
            self._set_value(.5)
        elif _min < value < _max:
            self._set_value(value)
        else:
            raise Exception('Slider value ({}) is not in range ({}-{})'.format(value, _min, _max))

    @property
    def range(self):
        """
        Range composed of min and max slider value.

        :rtype: ``collections.namedtuple``
        """
        return self._range

    @range.setter
    def range(self, range):
        default_range = (0.0, 1.0)
        _min, _max = default_range if range is None else range
        if _min > _max or _min == _max:
            raise Exception('Range min value has to be smaller than max value.')
        self._range = Range(_min, _max)
        self._set_range(Range(_min, _max))

    @property
    def on_slide(self):
        """
        The function for when the slider is slided

        :rtype:     ``callable``
        """
        return self._on_press

    @on_slide.setter
    def on_slide(self, handler):
        """
`       Set the function that is going to be executed on slide.

        :param handler:     The function to be executed
        :type handler:      ``callable`
        """
        self._on_press = handler
        self._set_on_slide(handler)

    def _set_on_slide(self, value):
        pass

    @property
    def enabled(self):
        """
        Indicates whether slider interaction is possible or not.

        :rtype:     ``Bool``
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        """
        Set the enabled state of the slider

        :param value:       The new enabled value
        :type value:        ``Bool``
        """
        if value is True:
            self._enabled = True
        elif value is False:
            self._enabled = False
        else:
            raise Exception('value must be of type Bool')
        self._set_enabled(value)



from .base import Widget


class Switch(Widget):
    """
    Switch widget, a clickable button with two stable states, True (on, checked) and False (off, unchecked)

    :param label:       Text to be shown next to the switch
    :type label:        ``str``

    :param id:          An identifier for this widget.
    :type  id:          ``str``

    :param style:       an optional style object. If no style is provided then a
                        new one will be created for the widget.
    :type style:        :class:`colosseum.CSSNode`

    :param on_toggle:    Function to execute when pressed
    :type on_toggle:     ``callable``

    :param is_on        Current on or off state of the switch
    :type is_on         ``Bool`
    """

    def __init__(self, label, id=None, style=None, on_toggle=None, is_on=False, enabled=True):
        super().__init__(id=id, style=style, label=label, on_toggle=on_toggle, is_on=is_on, enabled=enabled)

    def _configure(self, label, on_toggle, is_on, enabled):
        self.label = label
        self.on_toggle = on_toggle
        self.is_on = is_on
        self.enabled = enabled

    @property
    def label(self):
        """
        :returns: The label value
        :rtype: ``str``
        """
        return self._label

    @label.setter
    def label(self, value):
        """
        Set the label value

        :param value: The new label value
        :type  value: ``str``
        """
        if value is None:
            self._label = ''
        else:
            self._label = str(value)
        self._set_label(value)
        self.rehint()

    def _set_label(self, value):
        raise NotImplementedError('The inheriting class of {} must define _set_label()'.format(__class__))

    @property
    def on_toggle(self):
        """
        The callable function for when the switch is pressed

        :rtype: ``callable``
        """
        return self._on_toggle

    @on_toggle.setter
    def on_toggle(self, handler):
        self._on_toggle = handler
        self._set_on_toggle(handler)

    def _set_on_toggle(self, value):
        pass

    @property
    def is_on(self):
        """
        :returns: The is_on value

        :rtype: ``Bool``
        """
        return self._get_is_on()

    @is_on.setter
    def is_on(self, value):
        """
        Set the is_on value

        :param value: The new is_on value
        :type  value: ``Bool``
        """
        if value is True:
            self._set_is_on(True)
        elif value is False:
            self._set_is_on(False)

    @property
    def enabled(self):
        """
        :returns: The enabled state of the switch

        :rtype  value: ``Bool``
        """
        return self._get_enabled()

    @enabled.setter
    def enabled(self, value):
        """
        Set the enabled state of the switch

        :param value: The new enabled value
        :type  value: ``Bool``
        """
        if value is True:
            self._set_enabled(True)
        elif value is False:
            self._set_enabled(False)

    def _set_is_on(self, value):
        raise NotImplementedError('The inheriting class of {} must define _set_is_on()'.format(__class__))

    def _get_is_on(self):
        raise NotImplementedError('The inheriting class of {} must define _get_is_on()'.format(__class__))

    def _set_enabled(self, value):
        raise NotImplementedError('The inheriting class of {} must define _set_enabled()'.format(__class__))

    def _get_enabled(self):
        raise NotImplementedError('The inheriting class of {} must define _get_enabled()'.format(__class__))



from .base import Widget


class MultilineTextInput(Widget):
    '''
    A multi-line text input widget
    '''
    def __init__(self, id=None, style=None, initial=None):
        '''
        Instantiate a new instance of the multi-line text input widget

        :param id:          An identifier for this widget.
        :type  id:          ``str``

        :param style:       an optional style object. If no style is provided then a
                            new one will be created for the widget.
        :type style:        :class:`colosseum.CSSNode`

        :param initial:     The initial value
        :type  initial:     ``str``
        '''
        super().__init__(id=id, style=style, initial=initial)

    def _configure(self, initial):
        self.value = initial

    @property
    def value(self):
        '''
        The value of the multiline text input field

        :rtype: ``str``
        '''
        return self._value

    @value.setter
    def value(self, value):
        self._value = '' if value is None else str(value)
        self._set_value(self._value)
        self.rehint()

    def clear(self):
        '''
        Clear the value
        '''
        self.value = ''

    def _set_value(self, value):
        raise NotImplementedError('MultilineTextInput widget must define _set_value()')
